import enum
import io
import os
import re
import urllib
import warnings
from pathlib import Path
from typing import List, Optional, Tuple, Union, Type, IO, Callable, cast
from functools import wraps, partial

import gpg_lite as gpg
from libbiomedit import crypt
from .error import UserError
from .request import urlopen
from .secret import Secret

GPGStore = gpg.GPGStore
extract_pub_key_ids = gpg.extract_key_id
get_default_gpg_config_dir = gpg.get_default_gpg_config_dir

ExceptionType = Union[Type[BaseException], Tuple[Type[BaseException], ...]]


def to_user_error(error_types: ExceptionType):
    """A decorator to turn errors of type :error_types: into UserError"""

    def _to_user_error(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except error_types as e:
                raise UserError(format(e)) from e

        return wrapped

    return _to_user_error


def to_user_error_gen(error_types: ExceptionType):
    """A decorator to turn errors of type :error_types: into UserError"""

    def _to_user_error(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            try:
                yield from f(*args, **kwargs)
            except error_types as e:
                raise UserError(format(e)) from e

        return wrapped

    return _to_user_error


validate_pub_key = to_user_error(RuntimeError)(crypt.validate_pub_key)
validated_keys_by_ids = to_user_error_gen((RuntimeError, gpg.cmd.GPGError))(
    partial(crypt.validated_keys_by_ids, url_opener=urlopen)
)
refresh_keys = to_user_error((RuntimeError, gpg.cmd.GPGError))(
    partial(crypt.refresh_keys, url_opener=urlopen)
)
fingerprint2keyid = to_user_error(RuntimeError)(crypt.fingerprint2keyid)
verify_metadata_signature = to_user_error(RuntimeError)(
    partial(crypt.verify_metadata_signature, url_opener=urlopen)
)


@to_user_error(gpg.cmd.GPGError)
def search_keyserver(name: str, keyserver: str) -> List[gpg.KeyInfo]:
    """Search keyserver for keys maching the query"""
    return list(gpg.search_keys(name, keyserver, url_opener=urlopen))


@to_user_error(gpg.cmd.GPGError)
def import_keys(key_data: str, gpg_store: GPGStore):
    """Import keys from text"""
    return gpg_store.import_file(key_data.encode())


@to_user_error(gpg.cmd.GPGError)
def delete_keys(fingerprints: List[str], gpg_store: GPGStore):
    """Delete public key"""
    return gpg_store.delete_keys(*fingerprints)


@to_user_error(gpg.cmd.GPGError)
def create_revocation_certificate(
    fingerprint: str, passphrase: str, gpg_store: GPGStore
) -> bytes:
    """Create a revocation certificate for the key"""
    return gpg_store.gen_revoke(fingerprint, passphrase=passphrase)


def verify_key_length(key: gpg.Key, min_key_length: int = 4096):
    if key.key_length < min_key_length:
        raise UserError(
            f"Key {key.key_id} is shorter than the minimal "
            f"required length of {min_key_length} bit"
        )


def upload_keys(fingerprints: List[str], keyserver: str, gpg_store: GPGStore):
    """Upload public key to keyserver"""
    try:
        gpg_store.send_keys(*fingerprints, keyserver=keyserver, url_opener=urlopen)
    except urllib.error.URLError:
        raise KeyserverError(action="upload", keyserver=keyserver) from None


def download_keys(fingerprints: List[str], keyserver: str, gpg_store: GPGStore):
    """Download public key to keyserver"""
    try:
        gpg_store.recv_keys(*fingerprints, keyserver=keyserver, url_opener=urlopen)
    except urllib.error.URLError:
        raise KeyserverError(action="download", keyserver=keyserver) from None


def detach_sign_file(
    src: Union[bytes, IO[bytes]],
    signature_fingerprint: str,
    passphrase: str,
    gpg_store: GPGStore,
) -> bytes:
    """Sign a file with a detached signature"""
    try:
        with gpg_store.detach_sig(src, passphrase, signature_fingerprint) as out:
            return out.read()
    except gpg.cmd.GPGError as e:
        raise UserError(f"File signing failed. {e}") from e


def check_password(user_fingerprint: str, passphrase: str, gpg_store: GPGStore):
    """Check if key password is correct"""
    try:
        with gpg_store.detach_sig(b"test", passphrase, user_fingerprint):
            pass
    except gpg.cmd.GPGError:
        raise UserError("GPG password is incorrect") from None


class KeyType(enum.Enum):
    public = enum.auto()
    secret = enum.auto()


def open_gpg_dir(gpg_dir: str) -> GPGStore:
    """Open the database inside a GnuPG directory and return it as a gpg
    object.

    :param gpg_dir: path of the GnuPG directory to open.
    :return: a gnupg GPGStore
    :raises UserError:
    """
    if not Path(gpg_dir).is_dir():
        os.makedirs(gpg_dir, mode=0o700)

    try:
        return GPGStore(config_dir=gpg_dir)
    except ValueError:
        raise UserError(f"unable to open GnuPG directory [{gpg_dir}].") from None


def retrieve_keys(
    gpg_store: GPGStore,
    sender: str,
    recipients: List[str],
    validation_keyid: Optional[str] = None,
    keyserver_url: Optional[str] = None,
    with_refresh_keys: bool = True,
) -> Tuple[gpg.Key, List[gpg.Key]]:
    """Retrieve the following keys from the user's local GnuPG keyring:
     - public key of sender.
     - public key of recipient(s).
    Exactly one key must be found for each sender and recipient, otherwise an
    error is raised. A number of checks are also performed on the keys. The
    keys are then returned as a tuple with one sender key, and one or more
    recipient keys: (sender key, [recipient keys]).

    :param gpg_store: local GnuPG database as gpg-lite object.
    :param sender: fingerprint or email of data sender.
    :param recipients: fingerprint or email of data recipient(s). There can
        be one or more recipients.
    :param validation_keyid: fingerprint of validation key. Used to check that
        the retrieved key is signed by the validation key.
    :param keyserver_url: url of keyserver. Used for checking key origin.
    :param with_refresh_keys: if True, the found public keys are re-downloaded
        from the keyserver to make sure they are up-to-date.
    :return: tuple with one sender key, and one or more recipient keys.
    :raises UnpackError:
    """
    # Search public keys for sender and recipient(s).
    pub_keys = [
        search_pub_key(search_term=x, gpg_store=gpg_store, sigs=True)
        for x in [sender] + recipients
    ]

    # If asked to do so, refresh keys by re-downloading them from keyserver.
    if with_refresh_keys:
        pub_keys = refresh_keys(
            keys=pub_keys, gpg_store=gpg_store, sigs=True, keyserver_url=keyserver_url
        )
    # Verify keys are signed and non-revoked.
    for key in pub_keys:
        validate_pub_key(key, validation_keyid, keyserver_url)

    # Split keys into sender and recipient(s) keys.
    sender_key, *recipient_keys = pub_keys

    # Verify a private key matching the user's public key exists. Note that
    # the key itself is not really needed because it shares the fingerprint
    # with the public key, but it has to be there.
    search_priv_key(search_term=sender_key.fingerprint, gpg_store=gpg_store)

    # Return the sender's public key, and one or more recipient public keys
    # as two separate objects.
    return (sender_key, recipient_keys)


def search_pub_key(search_term: str, gpg_store: GPGStore, sigs: bool = True) -> gpg.Key:
    """Search for a single public key in the local keyring based on the input
    search term. If no or more than one key is matching the search term, an
    error is raised.

    :param gpg_store: key database as gnupg object.
    :param search_term: search term for the key, e.g. fingerprint or key owner
        email address.
    :param sigs: if True, return key with signatures.
    :return: GnuPG key matching the search term.
    :raises UnpackError:
    """
    keys = gpg_store.list_pub_keys(keys=(search_term,), sigs=sigs)
    try:
        (key,) = keys
    except ValueError:
        raise UnpackError(
            keys, key_type=KeyType.public, search_value=search_term
        ) from None
    return key


def search_priv_key(search_term: str, gpg_store: GPGStore) -> gpg.Key:
    """Searches the user's local keyring for a secret key matching the search
    term. Raises an error if no or more than one key are found.
    """
    keys = gpg_store.list_sec_keys(keys=(search_term,))
    try:
        (key,) = keys
    except ValueError:
        raise UnpackError(
            keys, key_type=KeyType.secret, search_value=search_term
        ) from None
    return key


class UnpackError(UserError):
    """Error class that displays an error message for the cases when a search
    for a public/secret key on the user's local keyring does not yield exactly
    one match"""

    def __init__(self, keys, key_type: KeyType, search_value: str):
        if not keys:
            msg_start = f"No {key_type.name} key"
        else:
            msg_start = f"Multiple {key_type.name} keys"
        super().__init__(msg_start + f" matching: {search_value}")


def assert_key_is_not_revoked(key: gpg.Key):
    """Check if a GnuPG key is revoked

    :param key: Key to check
    :raises UserError:
    """
    if key.validity is gpg.Validity.revoked:
        raise UserError(f"{key.uids[0]} key has been revoked")


def encrypt_and_sign(
    source: Callable,
    output: Callable,
    gpg_store: GPGStore,
    recipients_fingerprint: List[str],
    signature_fingerprint: Optional[str] = None,
    passphrase: Optional[str] = None,
    always_trust: bool = True,
):
    """Encrypt input data with a GnuPG public key and sign it with a GnuPG
    private key.

    In this function, the compression level of the "encrypt()" method is set
    to 0 as we are only encrypting data that has already been compressed, or
    has explicitely been requeseted by the user not to be compressed.

    There is no check of the validity of the sender and recipient keys, as it
    is assumed that this is already done earlier.

    :param source: callable writing data to encrypt.
    :param output: callable reading encrypted data.
    :param gpg_store: directory containing GnuPG keyrings as gnupg object.
    :param recipients_fingerprint: fingerprint of public key(s) with which
        the data should be encrypted.
    :param signature_fingerprint: fingerprint of private key with which the
        data should be signed. If the parameter value is set to None the
        encrypted file is not signed.
    :param passphrase: password associated to 'signature_fingerprint'. This
        parameter value can be set to None if the encrypted file should not be
        signed.
    :param always_trust: if False, the encryption key must be signed by the
        local user.
    :raises UserError:
    """
    try:
        # Note: gpg.CompressAlgo.NONE evaluates to "uncompressed", which
        # tells GnuPG to not compress the input data.
        gpg_store.encrypt(
            source=source,
            recipients=recipients_fingerprint,
            output=output,
            passphrase=passphrase,
            trust_model=gpg.TrustModel.always if always_trust else gpg.TrustModel.pgp,
            sign=signature_fingerprint,
            compress_algo=gpg.CompressAlgo.NONE,
        )
    except gpg.cmd.GPGError:
        raise UserError(
            "Encryption failed. " "Maybe you entered a wrong passphrase."
        ) from None


def get_recipient_email(key: gpg.Key) -> str:
    try:
        email = key.uids[0].email
    except (IndexError, UnpackError):
        raise UserError(
            f"Could not determine email address " f"for GPG key {key.key_id}."
        ) from None
    if not email:
        raise UserError(
            f"GPG key [{key.key_id}] does not contain a valid " f"email address."
        )
    return email


def decrypt(
    source: Union[io.FileIO, Callable],
    output: Union[str, Callable],
    gpg_store: GPGStore,
    passphrase: Optional[str],
) -> List[str]:
    """Decrypt data.

    :param source: data to encrypt.
    :param output: encrypted data.
    :param gpg_store: directory containing GnuPG keyrings as gnupg object.
    :param passphrase: password associated to 'signature_fingerprint'. This
        parameter value can be set to None if the encrypted file should not be
        signed.
    :return: fingerprint or keyid of the signee's key.
    :raises UserError:
    """
    try:
        if isinstance(output, str):
            with open(output, "wb") as fout:
                sender_fprs = gpg_store.decrypt(
                    source=source, output=cast(io.FileIO, fout), passphrase=passphrase
                )
        else:
            sender_fprs = gpg_store.decrypt(
                source=source, output=output, passphrase=passphrase
            )
    except gpg.cmd.GPGError as e:
        raise UserError(
            "Failed to decrypt. Error message from gpg:\n" + format(e)
        ) from e
    if not sender_fprs:
        warnings.warn("Encrypted package is not signed by the sender")
    elif len(sender_fprs) > 1:
        warnings.warn(
            "Encrypted package has multiple signatures. "
            "This is not compliant with the BiomedIT Protocol"
        )
    return sender_fprs


def create_key(
    full_name: str,
    email: str,
    pwd: Secret,
    pwd_repeat: Secret,
    gpg_store: GPGStore,
    key_type: str = "RSA",
    key_length: int = 4096,
) -> gpg.Key:
    """Create a new public/private key"""
    min_pwd_len = 10
    if len(full_name) < 5:
        raise UserError("Full name must be at least 5 characters long.")
    if not re.search(r"[^@]+@[^@]+\.[^@]+", email):
        raise UserError("Invalid email address.")
    if pwd != pwd_repeat:
        raise UserError("Password do not match.")
    if len(pwd) < min_pwd_len:
        raise UserError("Password is too short (min length: " f"{min_pwd_len})")
    fingerprint = gpg_store.gen_key(
        key_type=key_type,
        key_length=key_length,
        full_name=full_name,
        email=email,
        passphrase=pwd,
    )
    pkey = gpg_store.list_sec_keys((fingerprint,))
    if not pkey:
        raise UserError(f"No private keys found for: {fingerprint}")
    if len(pkey) > 1:
        raise UserError(f"Multiple private keys found for: {fingerprint}")
    return pkey[0]


class KeyserverError(UserError):
    """Error class that displays an error message when the keyserver cannot
    be reached.
    :param action: either "download" or "upload".
    :param keyserver: URL of keyserver.
    """

    def __init__(self, action: str, keyserver: str):
        super().__init__(
            f"Key {action} failed as sett could not connect to the keyserver.\n"
            f"A possible cause for this is that the keyserver is temporarily "
            f"unavailable.\n"
            f"Please try to connect to [{keyserver}] using your web browser to "
            f"confirm whether the keyserver is available or not.\n"
            f"If the issue persists, please contact the DCC."
        )



from typing import Sequence

from ..core.crypt import (
    upload_keys as crypt_upload_keys,
    verify_key_length,
    search_pub_key,
)
from ..core.error import UserError
from ..utils.log import create_logger
from ..utils.config import Config


logger = create_logger(__name__)


def upload_keys(key_ids: Sequence[str], *, config: Config):
    """Upload keys"""
    if config.offline:
        raise UserError("Uploading keys is not possible in the offline mode.")
    if config.keyserver_url is None:
        raise UserError("Keyserver URL is undefined.")
    keys = frozenset(search_pub_key(k, config.gpg_store, sigs=False) for k in key_ids)
    for key in keys:
        verify_key_length(key)
    if keys:
        logger.info("Uploading keys '%s'", ", ".join(k.key_id for k in keys))
        crypt_upload_keys(
            [k.fingerprint for k in keys],
            keyserver=config.keyserver_url,
            gpg_store=config.gpg_store,
        )



from pathlib import Path
from typing import Iterable, List

from sett import APP_NAME_SHORT
from .component import (
    SelectionAction,
    MandatoryLabel,
    show_warning,
    ToolBar,
    Action,
)
from .pyside import QtWidgets, QtCore, QtGui, QAction


class FileSelectionWidget(QtWidgets.QGroupBox):
    """File selection widget"""

    def __init__(self, label_text, parent, name_filter=None):
        super().__init__(parent)
        self.path = str(Path.home())
        self.name_filter = name_filter

        self.file_list_model = QtCore.QStringListModel()
        self.file_list_view = QtWidgets.QListView(self)
        self.file_list_view.setModel(self.file_list_model)
        self.file_list_view.setSelectionMode(
            QtWidgets.QAbstractItemView.ExtendedSelection
        )
        self.file_list_view.setLayout(QtWidgets.QVBoxLayout())

        toolbar = ToolBar("Add files", self)
        for action in self._create_actions():
            toolbar.addAction(action)
        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(MandatoryLabel(label_text))
        layout.addWidget(toolbar)
        layout.addWidget(self.file_list_view)

        self.setAcceptDrops(True)

    # pylint: disable=no-self-use
    def dragEnterEvent(self, event: QtGui.QDragEnterEvent) -> None:
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event: QtGui.QDropEvent) -> None:
        self._update_paths(url.toLocalFile() for url in event.mimeData().urls())

    def get_list(self) -> List[str]:
        """Returns the paths."""
        return self.file_list_model.stringList()

    def _create_actions(self) -> List[QtWidgets.QAbstractButton]:
        return [
            self._create_add_files_action(),
            self._create_remove_selected_action(),
            self._create_clear_list_action(),
        ]

    def _create_add_files_action(self) -> QAction:
        action = Action(QtGui.QIcon(":icon/feather/file-plus.png"), "Add file", self)
        action.triggered.connect(self._add_files)
        return action

    def _create_remove_selected_action(self) -> QAction:
        def clear_selected() -> None:
            indices = self.file_list_view.selectedIndexes()
            for index in indices:
                self.file_list_model.removeRows(index.row(), 1)
            self.file_list_model.layoutChanged.emit()

        action = SelectionAction(
            QtGui.QIcon(":icon/feather/file-minus.png"),
            "Remove selected",
            self,
            selection_model=self.file_list_view.selectionModel(),
        )
        action.triggered.connect(clear_selected)
        return action

    def _create_clear_list_action(self) -> QtWidgets.QAbstractButton:
        def clear_list() -> None:
            # Clear the selection BEFORE resetting the model
            self.file_list_view.selectionModel().clearSelection()
            self.file_list_model.setStringList([])
            self.file_list_model.layoutChanged.emit()

        action = Action(QtGui.QIcon(":icon/feather/trash-2.png"), "Clear list", self)
        action.triggered.connect(clear_list)
        return action

    def _update_paths(self, paths: Iterable[str]) -> None:
        paths = set(filter(None, paths))
        if paths:
            self.path = Path(next(iter(paths))).parent
        self.file_list_model.setStringList(
            sorted(set(self.file_list_model.stringList()) | paths)
        )
        self.file_list_model.layoutChanged.emit()

    def _add_files(self):
        dialog = QtWidgets.QFileDialog(self)
        dialog.setFileMode(QtWidgets.QFileDialog.ExistingFiles)
        dialog.setDirectory(str(self.path))
        if self.name_filter:
            dialog.setNameFilter(self.name_filter)
            dialog.selectNameFilter(self.name_filter)
        if dialog.exec_() == QtWidgets.QDialog.Accepted:
            self._update_paths(dialog.selectedFiles())


class DirectoryAndFileSelectionWidget(FileSelectionWidget):
    """File selection widget extension adding directory selection"""

    def __init__(self, label_text, parent):
        super().__init__(label_text, parent)

    def _create_actions(self):
        actions = super()._create_actions()
        actions.insert(1, self._create_add_directory_button())
        return actions

    def _create_add_directory_button(self):
        def add_directory():
            directory = QtWidgets.QFileDialog.getExistingDirectory(
                self, "Select directory", str(self.path)
            )
            self._update_paths((directory,))

        action = Action(
            QtGui.QIcon(":icon/feather/folder-plus.png"), "Add directory", self
        )
        action.triggered.connect(add_directory)
        return action


class ArchiveOnlyFileSelectionWidget(FileSelectionWidget):
    """File selection widget extension for selecting TAR archives only."""

    def __init__(self, label_text, parent):
        super().__init__(label_text, parent, "Archives *.tar (*.tar)")

    def dropEvent(self, event: QtGui.QDropEvent) -> None:
        paths = [
            url.toLocalFile()
            for url in event.mimeData().urls()
            if url.toLocalFile().endswith(".tar")
        ]
        if paths:
            self._update_paths(paths)
        else:
            show_warning(
                APP_NAME_SHORT,
                "Failed to load files. Only '.tar' archives are allowed.",
                self,
            )



import os
import json
from typing import Iterable, Tuple, Sequence, Optional
from datetime import date
from datetime import timedelta
from dataclasses import dataclass
import urllib.parse
import urllib.request
import base64

from .multipart import encode as multipart_encode
from . import protocol
from ..utils.progress import (
    progress_iter,
    progress_iter_with_sub_progress,
    ProgressInterface,
)
from ..core.error import UserError
from ..core.request import build_opener
from ..core.secret import Secret


@dataclass
class Protocol(protocol.Protocol):
    host: str
    api_key: Secret
    recipients: Sequence[str] = ()
    subject: str = "BiomedIT file transfer"
    message: str = ""
    expires_at: Optional[date] = None
    send_email: bool = False
    progress: Optional[ProgressInterface] = None
    chunk_size: int = 1024

    def upload(
        self,
        files: Sequence[str],
        progress: ProgressInterface = None,
    ):
        if self.expires_at is None:
            expires_at = date.today() + timedelta(days=7)

        attachment_ids = [
            upload_attachment(
                self.host,
                self.api_key,
                file_name,
                progress=sub_progress,
                chunk_size=self.chunk_size,
            )
            for file_name, sub_progress in progress_iter_with_sub_progress(
                files, progress
            )
        ]

        msg = {
            "message": {
                "recipients": self.recipients,
                "subject": self.subject,
                "message": self.message,
                "expires_at": expires_at.strftime("%Y-%m-%d"),
                "send_email": self.send_email,
                "authorization": 3,
                "attachments": attachment_ids,
            }
        }
        return post(
            self.host + "message",
            json.dumps(msg).encode(),
            self.api_key,
            "application/json",
        )


def upload_attachment(
    host: str,
    api_key: Secret,
    file_path: str,
    chunk_size: int = 1024,
    progress: ProgressInterface = None,
):
    with open(file_path, "rb") as stream:
        file_size = stream.seek(0, 2)
        stream.seek(0)
        n_chunks = 1 + (file_size - 1) // chunk_size
        file_name = os.path.basename(file_path)
        if n_chunks == 1:
            post_data: Iterable[Tuple[str, bytes]] = (
                multipart_encode("Filedata", file_name, stream.read()),
            )
        else:
            post_data = (
                multipart_encode(
                    "Filedata",
                    file_name,
                    chunk,
                    extra_data=(
                        ("name", file_name),
                        ("chunk", chunk_no),
                        ("chunks", n_chunks),
                    ),
                )
                for chunk_no, chunk in enumerate(chunk_iter(stream, chunk_size))
            )
        for content_type, data in progress_iter(
            post_data, progress, iterable_length=n_chunks
        ):
            attachment_id = post(host + "attachments", data, api_key, content_type)
    if not isinstance(attachment_id, bytes) or len(attachment_id) > 100:
        raise UserError(
            "LiquidFiles did not return a valid attachment_id. Something went wrong."
        )
    return attachment_id.decode()


def chunk_iter(stream, chunk_size: int):
    while True:
        chunk = stream.read(chunk_size)
        if not chunk:
            return
        yield chunk


def post(url: str, data: bytes, api_key: Secret, content_type: str):
    auth = "Basic " + base64.b64encode((api_key + ":x").encode()).rstrip().decode()
    request = urllib.request.Request(
        url,
        data,
        method="POST",
        headers={"Authorization": auth, "Content-Type": content_type},
    )
    with https_open(request) as response:
        response_text = response.read()
        return response_text


https_open = build_opener(
    default_opener=urllib.request.build_opener(
        urllib.request.HTTPSHandler(debuglevel=0)
    ).open
)



import os
import shutil
import warnings
from itertools import count
from pathlib import Path
from typing import List, Iterator, Union

from .error import UserError
from ..utils.log import create_logger

logger = create_logger(__name__)


def search_files_recursively(input_list: List[str]) -> Iterator[str]:
    """Recursively search for files and/or verify that files part of
    input_list really exist.

    :param input_list: list of files and directories to recursively search
        for files.
    :return: list of files and their absolute path.
    :raises UserError:
    """

    # Loop through all input path provided by the user. If the path is a
    # directory, search it recursively for files.
    for path in input_list:
        path_obj = Path(path).absolute()
        if path_obj.is_file():
            yield path_obj.as_posix()
        elif path_obj.is_dir():
            yield from (x.as_posix() for x in path_obj.rglob("*") if x.is_file())
        else:
            raise UserError(
                f"input path is not a valid file or directory: " f"{path_obj.name}"
            )


def check_file_read_permission(input_list: List[str]) -> None:
    """Verify the user has read permission on all files listed in input_list"""
    no_read_permission = [x for x in input_list if not os.access(x, os.R_OK)]
    if no_read_permission:
        raise UserError(
            f"no read permission on input files:" f"{', '.join(no_read_permission)}"
        )


def delete_files(*files: str) -> None:
    """Delete the specified file(s) and catch error if deletion fails.
    :param files: paths of files to delete from disk.
    :raises UserError:
    """
    failed = []
    for f in files:
        if os.path.exists(f):
            try:
                os.unlink(f)
            except BaseException:
                failed.append(f)

    if failed:
        warnings.warn(
            f"Failed to delete file(s): {', '.join(failed)}. "
            f"You have to take care of the clean up"
        )


def unique_filename(
    filename: str, extension: str = "", separator: str = "_", first_number: int = 1
):
    def filename_candidates():
        yield f"{filename}{extension}"
        yield from (
            f"{filename}{separator}{i}{extension}" for i in count(start=first_number)
        )

    return next(
        candidate
        for candidate in filename_candidates()
        if not os.path.exists(candidate)
    )


class DeleteDirOnError:
    """Context Manager to delete the specified directory including all of its files
    in case of an exception.
    """

    def __init__(self, directory: Union[str, Path]):
        self.directory = directory

    def __enter__(self):
        return self.directory

    def __exit__(self, exception_type, value, traceback):
        if traceback is not None:
            # Exception occurred
            if os.path.isdir(self.directory):
                logger.debug("Deleting directory: %s", self.directory)
                shutil.rmtree(self.directory)
            else:
                logger.debug("Directory doesn't exist: %s", self.directory)



from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from typing import (
    Any,
    Dict,
    Hashable,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
    Sequence,
    Callable,
)

from .pyside import QtCore
from ..core.metadata import Purpose
from ..utils.config import Config


class TableModel(QtCore.QAbstractTableModel):
    def __init__(
        self,
        data: Sequence[Sequence[Any]] = (),
        columns: Optional[Tuple[str, ...]] = None,
    ):
        super().__init__()
        self.set_data(data or tuple(tuple()))
        self.columns = columns or tuple(map(str, range(self.columnCount())))

    @property
    def columns(self) -> Tuple[str, ...]:
        return self._columns

    @columns.setter
    def columns(self, values: Sequence[str]):
        self._columns = tuple(values)

    def get_data(self) -> Tuple[Sequence[Any], ...]:
        return self._data

    def set_data(self, data: Sequence[Sequence[Any]]) -> None:
        self._data = tuple(sorted(tuple(x) for x in data))
        self.layoutChanged.emit()
        self.dataChanged.emit(
            self.createIndex(0, 0),
            self.createIndex(self.rowCount(), self.columnCount()),
        )

    def rowCount(self, _parent: QtCore.QModelIndex = QtCore.QModelIndex()) -> int:
        return len(self._data)

    def columnCount(self, _parent: QtCore.QModelIndex = QtCore.QModelIndex()) -> int:
        return len(self._data) and len(self._data[0])

    def headerData(
        self, section: int, orientation=QtCore.Qt.Horizontal, role=QtCore.Qt.DisplayRole
    ) -> Optional[str]:
        if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
            return self._columns[section]
        if orientation == QtCore.Qt.Vertical and role == QtCore.Qt.DisplayRole:
            return str(section)
        return None

    # Default value Qt.DisplayRole for index is not an int (upstream issue):
    def data(  # type: ignore
        self, index: QtCore.QModelIndex, role: int
    ) -> Optional[Hashable]:
        if role == QtCore.Qt.DisplayRole:
            return self._data[index.row()][index.column()]
        return None

    def removeRows(self, row: int, count: int, _parent=QtCore.QModelIndex()) -> bool:
        if row > -1 and row + count <= self.rowCount():
            self.beginRemoveRows(QtCore.QModelIndex(), row, row + count - 1)
            self.set_data(self.get_data()[:row] + self.get_data()[row + count :])
            self.endRemoveRows()
            return True
        return False


class KeyValueListModel(QtCore.QAbstractListModel):
    """List model extension for key-value objects."""

    def __init__(
        self, *args, data: Optional[Iterable[Tuple[Hashable, Any]]] = None, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.set_data(data or [])

    def set_data(self, data: Iterable[Tuple[Hashable, Any]]) -> None:
        self._keyvalues = dict(data)
        self._keys = list(self._keyvalues.keys())
        self.layoutChanged.emit()

    # Default value Qt.DisplayRole for index is not an int (upstream issue):
    def data(self, index, role) -> Optional[Hashable]:  # type: ignore
        if role == QtCore.Qt.DisplayRole:
            return self._keys[index.row()]
        return None

    def rowCount(self, _parent: QtCore.QModelIndex = QtCore.QModelIndex()) -> int:
        return len(self._keys)

    def get_value(self, index: Union[int, QtCore.QModelIndex]) -> Any:
        if isinstance(index, int):
            return self._keyvalues[self._keys[index]]
        if isinstance(index, QtCore.QModelIndex):
            return self._keyvalues[self._keys[index.row()]]
        raise TypeError("Wrong index type for index.")


@dataclass
class AppData:
    config: Config

    encrypt_sender: str = ""
    encrypt_recipients: Tuple[str, ...] = ()
    encrypt_transfer_id: Optional[str] = None
    encrypt_purpose: Optional[Purpose] = None
    encrypt_compression_level: int = 6
    encrypt_output_suffix: str = ""
    encrypt_output_location: Path = field(default_factory=Path.home)
    encrypt_files: List[str] = field(default_factory=list)
    encrypt_verify_dtr: bool = True

    decrypt_decrypt_only: bool = False
    decrypt_output_location: Path = field(default_factory=Path.home)
    decrypt_files: List[str] = field(default_factory=list)

    transfer_protocol_name: str = "sftp"
    transfer_protocol_args: Dict = field(default_factory=dict)
    transfer_files: List[str] = field(default_factory=list)
    transfer_verify_dtr: bool = True

    priv_keys_model: KeyValueListModel = field(default_factory=KeyValueListModel)
    pub_keys_model: KeyValueListModel = field(default_factory=KeyValueListModel)
    default_key_index: Optional[int] = None

    def __setattr__(self, attr, value):
        super().__setattr__(attr, value)
        if hasattr(self, "_listeners"):
            for callback in self._listeners[attr]:
                callback()

    def add_listener(self, attr, listener: Callable):
        self._listeners[attr].append(listener)

    def __post_init__(self):
        # pylint: disable=attribute-defined-outside-init
        self._listeners = defaultdict(list)
        if self.config.offline:
            self.encrypt_verify_dtr = self.transfer_verify_dtr = False
        if self.config.output_dir:
            self.encrypt_output_location = Path(self.config.output_dir)
            self.decrypt_output_location = Path(self.config.output_dir)
        if self.config.compression_level is not None:
            self.encrypt_compression_level = self.config.compression_level



import os
import sys
import json
import atexit
from logging import Logger
from functools import wraps
from contextlib import suppress, contextmanager
from datetime import datetime, timedelta
import warnings
from typing import Callable, Optional
from glob import glob

from .. import URL_GITLAB_ISSUES
from ..core.error import UserError
from ..core.filesystem import delete_files
from .config import Config, config_to_dict
from .log import log_to_memory

EXCEPTION_TYPES = (UserError, FileNotFoundError, FileExistsError, PermissionError)


def exit_on_exception(f: Callable) -> Callable:
    """A function decorator exiting the application on an exception of instance contained
    in or equal to `EXCEPTION_TYPES`, occurring while executing the decorated
    function. No stack trace will be printed, i.e. the errors are assumed to be handled /
    reported to the end users in a meaningful way.
    """

    @wraps(f)
    def wrapped_f(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except EXCEPTION_TYPES:
            sys.exit(1)

    return wrapped_f


def suppress_exceptions(f: Callable) -> Callable:
    """Decorator implemention of the contextlib.suppress context manager"""

    @wraps(f)
    def wrapped_f(*args, **kwargs):
        with suppress(*EXCEPTION_TYPES):
            return f(*args, **kwargs)

    return wrapped_f


ERROR_REPORT_PREFIX = "error_report-"


def write_error_report(config: Config, log: str, context: Optional[str] = None) -> str:
    """Writes a error report containing sys.argv, config and a log.
    Returns the filename of the report."""
    f_name = f"error_report-{datetime.now().strftime('%Y-%m-%dT%H-%M-%S')}.txt"
    with open(os.path.join(config.log_dir, f_name), "w") as f:
        f.write("cmd: " + repr(sys.argv) + "\n")
        if context is not None:
            f.write("context: " + context + "\n")
        f.write("---\nlog\n===\n")
        f.write(log)
        f.write("\n---\nconfig\n======\n")
        json.dump(config_to_dict(config), f, indent=2)
        return f.name


def cleanup_error_reports(log_dir: str):
    old_crash_reports = (
        f
        for f in glob(os.path.join(log_dir, ERROR_REPORT_PREFIX) + "*.txt")
        if datetime.now() - datetime.fromtimestamp(os.path.getmtime(f))
        > timedelta(weeks=1)
    )
    delete_files(*old_crash_reports)


def log_exceptions(logger: Logger) -> Callable[[Callable], Callable]:
    """A function decorator catching the first exception of instance contained
    in or equal to `EXCEPTION_TYPES`, occurring while executing the decorated
    function, logging an error and reraises the exception for further error handling.
    Also turns warnings into log entries
    """

    def decorator(f: Callable) -> Callable:
        @wraps(f)
        def wrapped_f(*args, **kwargs):
            try:
                with warnings.catch_warnings(record=True) as w:
                    warnings.simplefilter("always")
                    r = f(*args, **kwargs)
                    for warning in w:
                        logger.warning(format(warning.message))
                    return r
            except EXCEPTION_TYPES as e:
                logger.error(format(e))
                raise

        return wrapped_f

    return decorator


@contextmanager
def error_report_hint_at_exit(config: Config, **kwargs):
    """Displays a message about a generated error report as a last action before exit"""

    try:
        with error_report_on_exception(config, **kwargs):
            yield None
    except ExceptionWithReport as e:
        msg = e.error_report_msg
        atexit.register(lambda: print("\n" + msg, file=sys.stderr))
        if e.__cause__ is None:
            raise e
        raise e.__cause__


class ExceptionWithReport(Exception):
    """Exception class holding the additional info about the error report file"""

    def __init__(self, error_report_file: str):
        self.error_report_msg = (
            f"An error occurred. If you think it is as bug, please report it at "
            f"{URL_GITLAB_ISSUES} and attach the following auto generated error "
            f"report: {error_report_file}. "
            f"Or contact the developers via another channel (please also provide "
            f"the error report)."
        )
        self.error_report_file = error_report_file
        super().__init__(self.error_report_msg)

    def __format__(self, format_spec):
        return format(self.__cause__) + "\n" + self.error_report_msg


@contextmanager
def error_report_on_exception(config: Config, context: Optional[str] = None, **kwargs):
    """Function decorator / context manager generating a error report on any exception.

    In case of an exception, calls `handle_error_report` and reraises the exception
    for further handling.
    """
    with log_to_memory(**kwargs) as mem_log:
        try:
            yield None
        except Exception as e:
            error_report_file = write_error_report(
                config, mem_log.getvalue(), context=context
            )
            cleanup_error_reports(config.log_dir)
            raise ExceptionWithReport(error_report_file) from e



from functools import partial

from .component import TabMixin, PathInput, GuiProgress, run_dialog
from .file_selection_widget import ArchiveOnlyFileSelectionWidget
from .parallel import Worker
from .pyside import QtCore, QtWidgets
from ..workflows import decrypt


class DecryptTab(QtWidgets.QWidget, TabMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.threadpool = QtCore.QThreadPool.globalInstance()
        self.app_data = self.parent().app_data

        files_panel = self.create_files_panel()
        decrypt_options_panel = self.create_decrypt_options_panel()
        self.create_run_panel("Decrypt data", self.decrypt, "Decrypt selected files")
        self.app_data.add_listener("decrypt_files", self._enable_buttons)
        self.create_console()
        self.create_progress_bar()

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(files_panel)
        layout.addWidget(decrypt_options_panel)
        layout.addWidget(self.run_panel)
        layout.addWidget(self.console)
        layout.addWidget(self.progress_bar)

    def _enable_buttons(self):
        self.set_buttons_enabled(len(self.app_data.decrypt_files) > 0)

    def create_files_panel(self):
        box = ArchiveOnlyFileSelectionWidget(
            "Files to decrypt (hint: you can drag & drop files)", self
        )
        box.file_list_model.layoutChanged.connect(
            lambda: setattr(self.app_data, "decrypt_files", box.get_list())
        )
        return box

    def create_decrypt_options_panel(self):
        box = QtWidgets.QGroupBox("Output")

        decompress = QtWidgets.QCheckBox("Decompress", box)
        decompress.setStatusTip("Decompress data after decryption")
        decompress.setChecked(not self.app_data.decrypt_decrypt_only)
        decompress.stateChanged.connect(
            lambda state: setattr(
                self.app_data, "decrypt_decrypt_only", state == QtCore.Qt.Unchecked
            )
        )

        output_location = PathInput(path=self.app_data.encrypt_output_location)
        output_location.status_tip = "Destination folder for the decrypted packages"
        output_location.on_path_change(
            partial(setattr, self.app_data, "decrypt_output_location")
        )

        layout_output = QtWidgets.QHBoxLayout()
        layout_output.addWidget(QtWidgets.QLabel("Location"))
        layout_output.addWidget(output_location.text)
        layout_output.addWidget(output_location.btn)
        layout = QtWidgets.QVBoxLayout(box)
        layout.addWidget(decompress)
        layout.addLayout(layout_output)
        return box

    def decrypt(self, dry_run=False):
        progress = GuiProgress()
        progress.updated.connect(self.progress_bar.setValue)
        if not dry_run:
            pw = run_dialog(self, "Enter password for your GPG key")
            if pw is None:
                return
        else:
            pw = None
        worker = Worker(
            decrypt.decrypt,
            files=self.app_data.decrypt_files,
            capture_loggers=(decrypt.logger,),
            output_dir=str(self.app_data.decrypt_output_location),
            config=self.app_data.config,
            decrypt_only=self.app_data.decrypt_decrypt_only,
            passphrase=pw,
            dry_run=dry_run,
            progress=progress,
            ignore_exceptions=True,
            report_config=self.app_data.config,
        )
        self.add_worker_actions(worker)
        self.threadpool.start(worker)



#!/usr/bin/env python3

import os
import sys
from datetime import datetime
from setuptools import setup, find_packages

sys.path.insert(0, os.path.dirname(__file__))

# pylint: disable=wrong-import-position
from sett import URL_READTHEDOCS, URL_GITLAB

# 0.0.0-dev.* version identifiers for development only (not public)
__version__ = "0.0.0.dev" + datetime.now().strftime("%Y%m%d")

if __name__ == "__main__":
    setup(
        name="sett",
        version="2.8.1",
        license="LGPL3",
        description="Data compression, encryption and transfer tool",
        long_description=open("README.md").read(),
        long_description_content_type="text/markdown",
        author="Robin Engler, "
        "Jarosław Surkont, "
        "Gerhard Bräunlich, "
        "Kevin Sayers, "
        "Christian Ribeaud, "
        "François Martin",
        author_email="robin.engler@sib.swiss, "
        "jaroslaw.surkont@unibas.ch, "
        "gerhard.braeunlich@id.ethz.ch, "
        "sayerskt@gmail.com, "
        "christian.ribeaud@karakun.com, "
        "francois.martin@karakun.com",
        url="https://gitlab.com/biomedit/sett",
        install_requires=[
            "gpg-lite>=0.7.2, <0.8.0",
            "libbiomedit>=0.2.0, <0.3.0",
            "paramiko>=2.7.1",
            "PySide6>=6.0.1",
            "dataclasses ; python_version<'3.7'",
            "typing_extensions ; python_version<'3.8'",
        ],
        extras_require={
            "socks": ["PySocks"],
            "legacy": ["PySide2>=5.15.1"],
            "benchmark": ["pandas", "matplotlib", "seaborn"],
        },
        packages=find_packages(exclude=["test", "test.*", "benchmark"]),
        package_data={"sett": ["py.typed"]},
        zip_safe=False,
        python_requires=">=3.6",
        entry_points={
            "console_scripts": ["sett=sett.cli:run"],
            "gui_scripts": ["sett-gui=sett.gui:run"],
        },
        test_suite="test",
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
            "Operating System :: OS Independent",
        ],
        project_urls={
            "Documentation": URL_READTHEDOCS,
            "Source": URL_GITLAB,
        },
    )



import collections
import argparse
import inspect
import sys
from functools import wraps, partial as ft_partial
from typing import Generator, Union, Optional, Tuple
import pprint


class SubcommandBase:
    """Abstract definition for a single command.
    Accept by the Subcommands._add function"""

    def __init__(self, name, action, arguments, help=None):  # pylint: disable=redefined-builtin
        self.name = name
        self.help = help
        self.action = action
        self.arguments = arguments

    def add_to_parser(self, parser: argparse.ArgumentParser, actions):
        pass


class Subcommands:
    """Entry point to command line interfaces.

    Add subcommands by deriving from this class and specify
    commands in the static :subcommands: variable
    """
    description: Optional[str] = None
    required = True
    subcommands: Tuple[SubcommandBase, ...] = ()
    version: Optional[str] = None

    def __init__(self, *args, **kwargs):
        parser = argparse.ArgumentParser(
            description=self.description,
            formatter_class=argparse.RawDescriptionHelpFormatter)
        if self.version is not None:
            parser.add_argument('--version', action='version',
                                version=self.version)
        subparsers = parser.add_subparsers(
            dest='action', help='Action help')
        subparsers.required = self.required
        actions = {}
        for cmd in self.subcommands:
            cmd.add_to_parser(subparsers, actions)
        cmd_args = vars(parser.parse_args(*args, **kwargs))
        action = actions[cmd_args.pop("action")]
        action(**cmd_args)


class Subcommand(SubcommandBase):
    """A single subcommand corresponding to a function f of the main workflow
    """

    def __init__(self, f, **kwargs):
        omit = frozenset(getattr(f, "keywords", ()))
        omit_pos = frozenset(range(len(getattr(f, "args", ()))))
        args = arguments_by_signature(f,
                                      omit_positionals=omit_pos,
                                      omit_keywords=omit,
                                      **kwargs)
        super().__init__(name=kebab_case(f.__name__), action=f,
                         arguments=tuple(args), help=f.__doc__)

    def add_to_parser(self, parser, actions):
        subparser = parser.add_parser(self.name, help=self.help)
        actions[self.name] = self.action
        for arg in self.arguments:
            arg.add_to_parser(subparser)


class SubcommandGroup(SubcommandBase):
    """A subcommand consisting of multiple subsubcommands corresponding
    to functions without arguments"""

    def __init__(self, name, *fcts, help=None):  # pylint: disable=redefined-builtin
        def call(f):
            f()
        arguments = (Argument("--" + f.__name__, action="store_const",
                              dest="f", const=f) for f in fcts)
        super().__init__(name=name, action=call,
                         arguments=arguments, help=help)

    def add_to_parser(self, parser, actions):
        actions[self.name] = self.action
        subparser = parser.add_parser(self.name, help=self.help)
        group = subparser.add_mutually_exclusive_group(required=True)
        for arg in self.arguments:
            arg.add_to_parser(group)


class Argument:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def add_to_parser(self, parser):
        parser.add_argument(*self.args, **self.kwargs)

    @classmethod
    def primitive(cls, dest, arg_names, **kwargs):
        return (cls(*arg_names, dest=dest, **kwargs),)

    @classmethod
    def positional_bool(cls, dest, arg_names, **kwargs):
        return cls.primitive(dest, arg_names, type=parse_bool, **kwargs)

    @classmethod
    def keyword_bool(cls, dest, arg_names, default=False, help=None, **kwargs):  # pylint: disable=redefined-builtin
        kwargs.pop("required", None)
        long_arg_names = tuple(filter(lambda n: n.startswith("--"), arg_names))
        short_arg_names = tuple(
            filter(lambda n: n not in long_arg_names, arg_names))
        negation_flags = tuple(
            a.replace("--", "--no-") for a in long_arg_names)
        default_flags = arg_names if default else negation_flags
        if short_arg_names and default:
            raise ValueError(
                "booleans with short argument name and default=True are not supported")
        if help:
            help += " "
        else:
            help = ""
        help += f"(disable with {', '.join(negation_flags)}, default: {', '.join(default_flags)})"
        return ArgumentMutExGroup(
            cls(*arg_names,
                action="store_true",
                dest=dest,
                default=default,
                help=help,
                **kwargs),
            cls(*negation_flags,
                action="store_false",
                default=default,
                dest=dest,
                help=argparse.SUPPRESS
                )
        )


class ArgumentMutExGroup:
    def __init__(self, *args: Argument):
        self.args = args

    def add_to_parser(self, parser):
        grp = parser.add_mutually_exclusive_group()
        for arg in self.args:
            arg.add_to_parser(grp)


def decorate(f, *decorators):
    for dec in decorators:
        f = dec(f)
    return f


# Starting with python 3.7, the typing module has a new API
_origin_attr = "__extra__" if sys.version_info < (3, 7) else "__origin__"


def arguments_by_signature(f, overrides=None,
                           omit_keywords=frozenset(),
                           omit_positionals=frozenset()) -> Generator[Argument, None, None]:
    """Builds argparser arguments according to the signature of a function f.
    Type hint annotations are used to infer the type for argparse"""
    sig = inspect.signature(f)
    if overrides is None:
        overrides = {}
    unknown_parameters = set(overrides) - set(sig.parameters)
    if unknown_parameters:
        raise ValueError("Function {} does not have arguments {} "
                         "(which have been provided as parameter docs)".format(
                             f.__name__, ", ".join(unknown_parameters)))
    has_keyword_only = any(
        p.kind is inspect.Parameter.KEYWORD_ONLY
        for p in sig.parameters.values())
    for pos, (p_name, p) in enumerate(sig.parameters.items()):
        if p_name in omit_keywords or pos in omit_positionals:
            continue
        positional = has_keyword_only and \
            p.kind is not inspect.Parameter.KEYWORD_ONLY
        p_overrides = overrides.get(p_name, {})
        arg_names = make_arg_names(p_overrides.get(
            "name", kebab_case(p_name) if not positional else p_name),
            p_overrides.get("alias", ()), positional)
        default = p_overrides.get("default", p.default)
        args = dict(help=p_overrides.get("help"))
        t = p_overrides.get("type", p.annotation)
        if t is inspect.Signature.empty:
            raise ValueError(f"Argument {arg_names[0]} has no type annotation")
        opt_type = optional_type(t)
        if opt_type is not None:
            if default is inspect.Signature.empty:
                raise ValueError(
                    f"{f.__name__}: {p_name}: typing.Optional arguments are "
                    "allowed only when accompanied with a default value")
            t = opt_type

        if not positional:
            args["dest"] = p_overrides.get("dest", p_name)
            args.update(default_args(default))
        if t is bool:
            if positional:
                yield Argument.positional_bool(arg_names=arg_names,
                                               **args)
            else:
                yield Argument.keyword_bool(arg_names=arg_names, **args)
            continue
        if is_sequence(t):
            t = getattr(t, "__args__", (str,))[0]
            if default is not inspect.Signature.empty:
                args["default"] = default
            if positional:
                args["nargs"] = '+' if default is inspect.Signature.empty else '*'
            else:
                args["action"] = "append"
        yield Argument(*arg_names, type=t, **args)


def default_args(default):
    if default is inspect.Signature.empty:
        return {"required": True}
    return {"default": default}


def make_arg_names(name, aliases, positional: bool):
    """Combines name and aliases into one tuple of argument names
    including "--"
    """
    n_dashes = min(2, len(name))
    arg_prefix = "-" * n_dashes
    if positional:
        if aliases:
            raise ValueError(
                "Aliases not allowed for positional arguments")
        arg_prefix = ""
    arg_names = (arg_prefix + name,)
    if isinstance(aliases, str):
        aliases = (aliases,)
    return arg_names + tuple(aliases)


bool_true_literals = ("1", "true", "y", "yes")
bool_false_literals = ("0", "false", "n", "no")

bool_literal_mapping = dict(
    [(l, True) for l in bool_true_literals] +
    [(l, False) for l in bool_false_literals])


def parse_bool(s):
    return bool_literal_mapping[s]


def is_sequence(t):
    t_origin = getattr(t, "__origin__", None)
    return isinstance(t_origin, type) and \
        issubclass(t_origin, collections.abc.Sequence)


def partial(*args, **kwargs):
    """Modified version of :functools.partial: which preserves
    __name__ and __doc__ attributes and can be used as a decorator
    """
    def decorator(f):
        keywords = getattr(f, "keywords", {})
        f_new = ft_partial(f, *args, **{**keywords, **kwargs})
        f_new.__name__ = f.__name__
        f_new.__doc__ = f.__doc__
        return f_new
    return decorator


def lazy_partial(*args, **kwargs):
    """Similar to :functools.partial: but load bound arguments at runtime of
    the wrapped function, by calling the arguments.
    """
    def decorator(f):
        @wraps(f)
        def wrapped_f(*_args, **_kwargs):
            newkeywords = {key: val()
                           for key, val in kwargs.items()}
            newkeywords.update(_kwargs)
            return f(*map(lambda f: f(), args), *_args, **newkeywords)

        wrapped_f.args = getattr(f, "args", ()) + args
        new_keywords = getattr(f, "keywords", {})
        new_keywords.update(kwargs)
        wrapped_f.keywords = new_keywords

        return wrapped_f
    return decorator


def set_default(**defaults):
    """Sets or changes default values of the decorated function"""
    def decorator(f):
        @wraps(f)
        def _f(*args, **kwargs):
            return f(*args, **{**defaults, **kwargs})
        return _f
    return decorator


def rename(name: str):
    """Changes the name of a function"""
    def wrapper(f):
        def wrapped(*args, **kwargs):
            return f(*args, **kwargs)
        wrapped.__name__ = name
        return wrapped
    return wrapper


def block(*arg_names: str):
    """Hides argument of a function from arguments_by_signature"""
    def wrapper(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            return f(*args, **kwargs)
        keywords = getattr(wrapped, "keywords", {})
        keywords.update({n: object() for n in arg_names})
        wrapped.keywords = keywords
        return wrapped
    return wrapper


def return_to_stdout(f):
    """Converts a function :f: into a function pretty printing the
    return value of :f: and returning None
    """
    pp = pprint.PrettyPrinter(indent=2)

    @wraps(f)
    def wrapper(*args, **kwargs):
        pp.pprint(f(*args, **kwargs))
    return wrapper


def kebab_case(name):
    return name.replace("_", "-")


def optional_type(T: type) -> Optional[type]:
    """For Optional types provided by typing, return the underlying python type,
    else None.
    E.g. returns int for typing.Optional[int]"""
    if getattr(T, "__origin__", None) is not Union:
        return None
    try:
        # Here T has attribute __origin__. Thus it is a typing._GenericAlias
        T1, T2 = T.__args__  # type: ignore
    except ValueError:
        return None
    if isinstance(None, T1):
        return T2
    if isinstance(None, T2):
        return T1
    return None



from dataclasses import dataclass

from libbiomedit import metadata as _metadata

from .error import UserError

# Reexports
alnum_str = _metadata.alnum_str
METADATA_FILE = _metadata.METADATA_FILE
Purpose = _metadata.Purpose
HexStr1024 = _metadata.HexStr1024
HexStr256 = _metadata.HexStr256


@dataclass(frozen=True)
class MetaData(_metadata.MetaData):
    """Wrapper around libbiomedit.metadata.MetaData throwing UserError"""

    @classmethod
    def from_dict(cls, d: dict):
        try:
            return super().from_dict(d)
        except ValueError as e:
            raise UserError(format(e)) from e


load_metadata = MetaData.from_dict



import os
import urllib
import urllib.request

from .error import UserError


def build_opener(default_opener=urllib.request.urlopen):
    proxy_env = os.environ.get("HTTPS_PROXY") or os.environ.get("ALL_PROXY")
    if proxy_env is not None and proxy_env.startswith("socks"):
        protocol, host, port = proxy_env.split(":")
        return socks_opener((protocol, host.lstrip("/"), int(port)))
    return default_opener


def socks_opener(proxy_settings):
    """Returns an opener based on proxy settings: Normal opener if no proxy
    settings are given, PySocks opener if proxy settings are given
    """
    return urllib.request.build_opener(socks_handler(*proxy_settings)).open


def socks_handler(protocol, host, port):
    try:
        import socks  # pylint: disable=import-outside-toplevel

        # pylint: disable=import-outside-toplevel
        from sockshandler import (
            SocksiPyHandler,
        )
    except ModuleNotFoundError:
        raise UserError(
            "module PySocks needs to be available when using a proxy"
        ) from None
    protocol_mapping = {"socks4": socks.SOCKS4, "socks5": socks.SOCKS5}
    try:
        protocol = protocol_mapping[protocol]
    except KeyError:
        raise UserError("Invalid scheme " + protocol) from None
    return SocksiPyHandler(protocol, host, port)


urlopen = build_opener()


def post(url: str, data: bytes) -> bytes:
    request = urllib.request.Request(
        url,
        data,
        method="POST",
        headers={"Content-Type": "application/json; charset=UTF-8"},
    )
    with urlopen(request) as response:
        response_text = response.read()
        return response_text



import logging
import sys
import traceback
from typing import Sequence, Callable, Optional, cast
from contextlib import contextmanager

from .pyside import QtCore
from ..utils.log import LOG_FORMAT
from ..utils.error_handling import (
    suppress_exceptions,
    log_exceptions,
    error_report_on_exception,
)
from ..utils.config import Config
from ..core.secret import enforce_secret_by_signature

from .. import RUNTIME_INFO


class LoggerHandler(logging.Handler):
    def __init__(self, level=logging.NOTSET, callback=None):
        super().__init__(level)
        self.callback = callback

    def emit(self, record):
        if self.callback:
            self.callback(self.format(record))


class WorkerSignals(QtCore.QObject):
    started = QtCore.Signal()
    finished = QtCore.Signal()
    error = QtCore.Signal(tuple)
    result = QtCore.Signal(object)
    progress = QtCore.Signal(int)
    logging = QtCore.Signal(str)


class Worker(QtCore.QRunnable):
    """Worker can run an arbitrary function in a separate thread.

    :param fn: function to run in a thread
    :param args: arguments passed to the function
    :param kwargs: keyword arguments passed to the function
    :param logger: logs from those loggers are passed to the logging
                   signal
    """

    def __init__(
        self,
        fn,
        *args,
        capture_loggers: Sequence[logging.Logger] = (),
        ignore_exceptions: bool = False,
        forward_errors: Optional[Callable[[str], None]] = None,
        report_config: Config,
        **kwargs,
    ):
        super().__init__()
        logger = (
            capture_loggers[0] if capture_loggers else cast(logging.Logger, logging)
        )
        # Somehow pyling does not know, that @contextmanager also is a decorator:
        sanitized_args, sanitized_kwargs = enforce_secret_by_signature(fn, args, kwargs)
        workflow_args_info = ",\n  ".join(
            [repr(arg) for arg in sanitized_args]
            + [f"{key}={repr(val)}" for key, val in sanitized_kwargs.items()]
        )
        self.fn = log_exceptions(logger=logger)(fn)
        if report_config.error_reports:
            self.fn = error_report_on_exception(  # pylint: disable=not-callable
                report_config,
                context=f"{fn.__name__}({workflow_args_info})",
            )(self.fn)
        if ignore_exceptions:
            self.fn = suppress_exceptions(self.fn)
        self.args = args
        self.kwargs = kwargs
        self.capture_loggers = capture_loggers
        self.signals = WorkerSignals()  # progress signal is not implemented
        if forward_errors is not None:
            self.signals.error.connect(lambda err: forward_errors(err[1]))

    def run(self) -> None:
        with attach_loggers_to_signals(self.capture_loggers, self.signals):
            try:
                self.signals.started.emit()
                self.signals.logging.emit(RUNTIME_INFO)
                result = self.fn(*self.args, **self.kwargs)
            except Exception:  # pylint: disable=broad-except
                exctype, value, _ = sys.exc_info()
                self.signals.error.emit((exctype, value, traceback.format_exc()))
            else:
                self.signals.result.emit(result)
        self.signals.finished.emit()


@contextmanager
def attach_loggers_to_signals(loggers, signals):
    logger_handler = LoggerHandler(callback=signals.logging.emit)
    logger_handler.setFormatter(logging.Formatter(LOG_FORMAT))
    logger_handler.setLevel(logging.INFO)
    for logger in loggers:
        logger.addHandler(logger_handler)
    yield None
    for logger in loggers:
        logger.removeHandler(logger_handler)



import re
import os
import io
from datetime import datetime
from dataclasses import dataclass
from typing import Optional, Callable, Tuple, Sequence
from pathlib import Path, PurePosixPath
from contextlib import contextmanager

from paramiko import (
    RSAKey,
    DSSKey,
    ECDSAKey,
    Ed25519Key,
    Transport,
    ssh_exception,
    SSHClient,
    AutoAddPolicy,
    SSHException,
    Agent,
)

from ..utils.progress import ProgressInterface
from ..utils.log import create_logger
from ..utils.config import get_config_file
from ..core.error import UserError
from ..core.secret import Secret
from .defs import ENVELOPE_DIR_FMT
from . import protocol


logger = create_logger(__name__)


@dataclass
class Protocol(protocol.Protocol):
    host: str
    username: str
    destination_dir: str
    two_factor_callback: Callable
    envelope_dir: Optional[str] = None
    pkey: Optional[str] = None
    pkey_password: Optional[Secret] = None
    pkey_password_encoding: str = (
        "utf_8"  # nosec (False Positive: pkey_password_encoding)
    )
    jumphost: Optional[str] = None

    def required_password_args(self) -> Sequence[str]:
        if self.pkey is not None and self.pkey_password is None:
            return ("pkey_password",)
        return ()

    def upload(
        self,
        files: Sequence[str],
        progress: Optional[ProgressInterface] = None,
    ):
        if self.envelope_dir is None:
            envelope_dir = datetime.now().strftime(ENVELOPE_DIR_FMT)
        progress_callback = progress and (lambda x, y: progress.update(x / y))
        remote_dir = PurePosixPath(self.destination_dir) / envelope_dir
        try:
            with sftp_connection(
                host=self.host,
                username=self.username,
                pkey=self.pkey,
                jumphost=self.jumphost,
                pkey_password=self.pkey_password,
                pkey_password_encoding=self.pkey_password_encoding,
                two_factor_callback=self.two_factor_callback,
            ) as sftp:
                try:
                    sftp.mkdir(str(remote_dir))
                except FileNotFoundError as e:
                    raise UserError(
                        "Remote destination directory does not exist: "
                        f"{self.destination_dir}"
                    ) from e
                except PermissionError as e:
                    raise UserError(
                        "You don not have enough permissions to write "
                        f"to the remote directory: {self.destination_dir}"
                    ) from e
                for tar in files:
                    remotepath = str(remote_dir / Path(tar).name)
                    remotepath_part = remotepath + ".part"
                    status = sftp.put(
                        localpath=os.path.realpath(tar),
                        remotepath=remotepath_part,
                        callback=progress_callback,
                        confirm=True,
                    )
                    remote_size = status.st_size
                    local_size = os.path.getsize(os.path.realpath(tar))

                    if local_size != remote_size:
                        raise UserError(
                            f"Incomplete file transfer: '{tar}'\n"
                            f"Remote: {remote_size}\nLocal: {local_size}"
                        )

                    try:
                        sftp.posix_rename(remotepath_part, remotepath)
                    except IOError as e:
                        raise UserError(format(e)) from e
                with io.BytesIO(b"") as fl:
                    sftp.putfo(
                        fl=fl,
                        remotepath=str(remote_dir / "done.txt"),
                        callback=progress_callback,
                        confirm=True,
                    )
        except ssh_exception.AuthenticationException as e:
            raise UserError(format(e)) from e


@contextmanager
def sftp_connection(  # nosec
    host: str,
    username: str,
    two_factor_callback: Callable,
    pkey: Optional[str] = None,
    pkey_password: Optional[Secret] = None,
    pkey_password_encoding: str = "utf_8",
    jumphost: Optional[str] = None,
):
    key = pkey and private_key_from_file(
        str(Path(pkey).expanduser()), pkey_password, encoding=pkey_password_encoding
    )
    if jumphost is not None:
        pw = two_factor_callback()
        sock = proxy_socket(host, jumphost, username, pkey=key, password=pw)
    else:
        sock = host
    trans = Transport(sock)
    trans.connect()
    try:
        auth(trans, username, key, two_factor_callback)
        sftp_client = None
        try:
            sftp_client = trans.open_sftp_client()
            yield sftp_client
        finally:
            if sftp_client:
                sftp_client.close()
    finally:
        trans.close()


def auth(trans, username, key, two_factor_callback: Callable):
    allowed_types = set()
    if key:
        allowed_types = trans.auth_publickey(username, key)
    else:
        try:
            allowed_types = auth_from_agent(trans, username)
        except SSHException:
            trans.auth_timeout = 120
            trans.auth_interactive(username, auth_handler)
    two_factor = bool(set(allowed_types) & _TWO_FACTOR_TYPES)
    if two_factor:
        f2_code = two_factor_callback()
        trans.auth_password(username, f2_code)


def proxy_socket(host, jumphost, username, **kwargs):
    tunnel = SSHClient()
    tunnel.set_missing_host_key_policy(AutoAddPolicy())
    tunnel.connect(
        jumphost,
        username=username,
        allow_agent=True,
        **kwargs,
    )
    return tunnel.get_transport().open_channel(
        "direct-tcpip", parse_host(host), parse_host(jumphost)
    )


def parse_host(host: str) -> Tuple[str, int]:
    try:
        _host, port = host.split(":")
        return _host, int(port)
    except ValueError:
        return host, 22


def auth_handler(_title, _instructions, prompt_list):
    if prompt_list:
        auth_url = re.findall(r"(https?://\S+)", prompt_list[0][0])
        if auth_url:
            logger.info("Authenticate at: %s", auth_url[0])
    resp = ["" for _ in prompt_list]
    return resp


def is_ascii(s):
    return all(ord(c) < 128 for c in s)


def private_key_from_file(path, password, encoding="utf_8"):
    errors = []
    pass_bytes = None if password is None else password.encode(encoding)
    for pkey_class in (RSAKey, DSSKey, ECDSAKey, Ed25519Key):
        try:
            return pkey_class.from_private_key_file(path, pass_bytes)
        except (SSHException, ValueError) as e:
            errors.append(e)
    if password is not None and not is_ascii(password):
        errors = [
            (
                "Your ssh secret key's password seems to contain "
                "some non-ascii characters.\n"
                "Either change your password ("
                "`ssh-keygen -f <path to your private key> -p`)"
                " or make sure the config option "
                "`ssh_password_encoding` is set to the same "
                "encoding, your key has been created with."
                "Your config file is here:\n"
                + get_config_file()
                + "\nThe encoding is usually `utf_8` on linux "
                "/ mac and `cp437` on windows for keys generated "
                "with ssh-keygen"
            )
        ] + errors
    raise UserError(
        "Could not load private key:\n" + "\n".join(format(e) for e in errors)
    )


def auth_from_agent(transport, username):
    agent = Agent()
    try:
        for key in agent.get_keys():
            try:
                logger.debug("Trying SSH agent key %s", key.get_fingerprint())
                # for 2-factor auth a successfully auth'd key password
                # will return an allowed 2fac auth method
                return transport.auth_publickey(username, key)
            except SSHException:
                pass
    finally:
        agent.close()
    raise SSHException("Could not load key from ssh agent")


_TWO_FACTOR_TYPES = {"keyboard-interactive", "password"}



import warnings

from sett import (
    APP_NAME_LONG,
    URL_READTHEDOCS,
    URL_GITLAB_ISSUES,
    APP_NAME_SHORT,
    __version__,
    URL_GITLAB,
    VERSION_WITH_DEPS,
)
from .component import show_warning, is_macos
from .decrypt_tab import DecryptTab
from .encrypt_tab import EncryptTab
from .keys_tab import KeysTab
from .model import AppData
from .parallel import Worker
from .pyside import QtCore, QtGui, QtWidgets, QAction
from .settings_tab import SettingsTab, config_to_settings
from .transfer_tab import TransferTab
from ..core.versioncheck import check_version
from ..utils.config import load_config, save_config, config_to_dict

QtCore.QThreadPool.globalInstance().setExpiryTimeout(-1)


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()
        self.title = f"{APP_NAME_LONG} ({__version__})"
        self.app_data = AppData(config=self.get_config())
        self.setWindowTitle(self.title)
        self.setMinimumSize(QtCore.QSize(900, 900))
        self.add_menu()
        self.add_tabs()
        self.add_status_bar()
        self.check_version()

    def get_config(self):
        with warnings.catch_warnings(record=True) as w:
            config = load_config()
            if w:
                show_warning(
                    "Configuration Error",
                    "\n".join(format(warning.message) for warning in w),
                    self,
                )
        return config

    def add_tabs(self):
        tab_keys = KeysTab(self)  # FIXME created first to populate gpg keys
        tab_encrypt = EncryptTab(self)
        tab_decrypt = DecryptTab(self)
        tab_transfer = TransferTab(self)
        tab_settings = SettingsTab(self)

        tabs = QtWidgets.QTabWidget()
        tabs.addTab(tab_encrypt, QtGui.QIcon(":icon/feather/lock.png"), "&Encrypt")
        tabs.addTab(tab_transfer, QtGui.QIcon(":icon/feather/send.png"), "&Transfer")
        tabs.addTab(tab_decrypt, QtGui.QIcon(":icon/feather/unlock.png"), "&Decrypt")
        tabs.addTab(tab_keys, QtGui.QIcon(":icon/feather/key.png"), "&Keys")
        tabs.addTab(
            tab_settings, QtGui.QIcon(":icon/feather/settings.png"), "&Settings"
        )

        self.setCentralWidget(tabs)

    def add_status_bar(self):
        self.status = QtWidgets.QStatusBar()
        self.setStatusBar(self.status)

    def add_menu(self):
        action_exit = QAction(QtGui.QIcon(":icon/feather/log-out.png"), "&Exit", self)
        action_exit.setShortcut(QtGui.QKeySequence("Ctrl+Q"))
        action_exit.setStatusTip("Exit application")
        action_exit.triggered.connect(self.close)

        menu = self.menuBar()
        menu.setNativeMenuBar(not is_macos())
        menu_file = menu.addMenu("&File")
        menu_file.addAction(action_exit)

        action_help = QAction(
            QtGui.QIcon(":icon/feather/book-open.png"), "&Documentation", self
        )
        action_help.setStatusTip("Open online documentation")
        action_help.setShortcut(QtGui.QKeySequence(QtGui.QKeySequence.HelpContents))
        action_help.triggered.connect(open_url(URL_READTHEDOCS))

        action_bug_report = QAction(
            QtGui.QIcon(":icon/feather/alert-triangle.png"), "&Report an Issue", self
        )
        action_bug_report.setStatusTip("Open online bug report form")
        action_bug_report.triggered.connect(open_url(URL_GITLAB_ISSUES))

        action_about = QAction(QtGui.QIcon(":icon/feather/info.png"), "&About", self)
        action_about.setStatusTip("Show info about application")
        action_about.triggered.connect(self.show_about)

        menu_help = menu.addMenu("&Help")
        menu_help.addAction(action_help)
        menu_help.addAction(action_bug_report)
        menu_help.addAction(action_about)

    def closeEvent(self, event):
        if config_to_settings(self.app_data.config) != config_to_settings(
            load_config()
        ):
            reply = QtWidgets.QMessageBox.question(
                self,
                "Persist changed settings?",
                "You made changes to settings you did not persist yet, "
                "do you want to permanently persist them to your sett configuration file?",
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                QtWidgets.QMessageBox.No,
            )
            if reply == QtWidgets.QMessageBox.Yes:
                save_config(config_to_dict(self.app_data.config))

        if self.app_data.config.gui_quit_confirmation:
            reply = QtWidgets.QMessageBox.question(
                self,
                "Quit",
                "Do you really want to quit?",
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                QtWidgets.QMessageBox.No,
            )
            if reply == QtWidgets.QMessageBox.Yes:
                event.accept()
            else:
                event.ignore()

    def check_version(self):
        if self.app_data.config.offline or not self.app_data.config.check_version:
            return

        def get_warnings():
            with warnings.catch_warnings(record=True) as w:
                check_version(self.app_data.config.repo_url, gui_formatting=True)
                return "\n".join(format(warning.message) for warning in w)

        worker = Worker(
            get_warnings,
            report_config=self.app_data.config,
        )
        worker.signals.result.connect(lambda x: x and show_warning(self.title, x, self))
        QtCore.QThreadPool.globalInstance().start(worker)

    def show_about(self):
        msg = QtWidgets.QMessageBox(parent=self)
        msg.setWindowTitle(f"About {APP_NAME_SHORT}")
        msg.setIcon(QtWidgets.QMessageBox.Information)
        msg.setTextFormat(QtCore.Qt.RichText)
        msg.setText(
            f"{APP_NAME_LONG}<br>"
            f"{VERSION_WITH_DEPS}<br><br>"
            f"For documentation go to "
            f"<a href='{URL_READTHEDOCS}'>{URL_READTHEDOCS}</a><br>"
            f"To report an issue go to "
            f"<a href='{URL_GITLAB_ISSUES}'>{URL_GITLAB_ISSUES}</a><br>"
            f"Source code is available at "
            f"<a href='{URL_GITLAB}'>{URL_GITLAB}</a><br><br>"
            f"{APP_NAME_SHORT} is developed as part of the "
            f"<a href='https://sphn.ch/network/projects/biomedit/'>BioMedIT "
            f"project</a>"
        )
        msg.show()


def open_url(url: str):
    """Returns a function that will opens the specified URL in the user's
    default browser when called. The function that is returned has no
    arguments.
    :param url: URL to open.
    :returns: function that opens specifed URL.
    """

    def open_url_template():
        if not QtGui.QDesktopServices.openUrl(QtCore.QUrl(url)):
            msg_warn = QtWidgets.QMessageBox()
            msg_warn.setWindowTitle("Warning")
            msg_warn.setText(f"Unable to open URL at \n{url}.")
            msg_warn.setIcon(QtWidgets.QMessageBox.Warning)
            msg_warn.exec_()

    return open_url_template



import sys
import time
import locale

from ..utils.progress import ProgressInterface


class CliProgress(ProgressInterface):
    """Terminal progress bar

    :param prefix: prefix string (Optional)
    :param suffix: suffix string (Optional)
    :param decimals: positive number of decimals in percent complete
    :param bar_length: character length of bar
    :param fill: Fill character for the bar
    :param initial_print: If true, do an intial print at :intital_fraction:
    :param initial_fraction: The initial position, the progress bar will start at
    """

    def __init__(
        self,
        prefix: str = "",
        suffix: str = "",
        decimals: int = 1,
        bar_length: int = 60,
        fill: str = "█",
        initial_print: bool = True,
        initial_fraction: float = 0.0,
    ):
        _, encoding = locale.getlocale()
        if encoding != "UTF-8" and fill == "█":
            fill = "#"

        self.total = 1.0
        self.n0 = initial_fraction
        self.n = initial_fraction
        self.prefix = prefix + " " if prefix else ""
        self.suffix = " " + suffix if suffix else ""
        self.decimals = decimals
        self.bar_length = bar_length
        self.fill = fill
        self.t0 = time.time()
        self.t = self.t0
        self.line_length = 0
        self._update_precision = decimals + 2
        if initial_print:
            self.print_current()

    def update(self, completed_fraction: float) -> None:
        t = time.time()
        completed_fraction_rounded = round(completed_fraction, self._update_precision)
        # != also handles reseting progress
        if completed_fraction_rounded != self.n or t - self.t > 1.0:
            self.n = completed_fraction_rounded
            self.t = t
            self.print_current()

    def get_completed_fraction(self) -> float:
        return self.n

    def print_current(self):
        percent = ("{0:." + str(self.decimals) + "f}").format(
            100 * (self.n / float(self.total))
        )
        filled_length = int(self.bar_length * self.n // self.total)
        char_bar = self.fill * filled_length + "-" * (self.bar_length - filled_length)
        if self.n == self.n0:
            eta_str = ""
        else:
            eta = self.t - self.t0
            if self.n != self.total:
                eta *= (self.total - self.n) / (self.n - self.n0)

            eta_str = " (" + format_eta(eta) + "s)"
        line = (
            "\r"
            + self.prefix
            + "|"
            + char_bar
            + "| "
            + percent
            + "%"
            + eta_str
            + self.suffix
        )
        fill_space = " " * max(0, self.line_length - len(line))
        self.line_length = len(line)

        sys.stdout.write(line + fill_space)
        if self.n == self.total:
            sys.stdout.write("\n")
        sys.stdout.flush()


def format_eta(eta_secs: float) -> str:
    eta_h, eta_secs = divmod(eta_secs, 3600)
    eta_min, eta_secs = divmod(eta_secs, 60)
    if eta_h != 0.0:
        return f"{int(eta_h):02d}:{int(eta_min):02d}:{int(eta_secs):02d}"
    if eta_min != 0.0:
        return f"{int(eta_min):02d}:{int(eta_secs):02d}"
    return str(int(eta_secs))



import os
import io
import tarfile
import time
from abc import ABC, abstractmethod
from pathlib import Path
from contextlib import contextmanager
from itertools import filterfalse
from typing import List, Iterable, IO, Sequence, Union

try:
    from typing import TypedDict
except ImportError:
    from typing_extensions import TypedDict

from .error import UserError
from .filesystem import delete_files
from ..utils.progress import ProgressInterface, with_progress, progress_iter


tarfile.DEFAULT_FORMAT = tarfile.PAX_FORMAT

METADATA_FILE = "metadata.json"
METADATA_FILE_SIG = "metadata.json.sig"
CHECKSUM_FILE = "checksum.sha256"
DATA_ARCHIVE = "data.tar.gz"
DATA_FILE_ENCRYPTED = DATA_ARCHIVE + ".gpg"
PACKAGE_CONTENT = {METADATA_FILE, METADATA_FILE_SIG, DATA_FILE_ENCRYPTED}
CONTENT_FOLDER = "content"


def check_tar(tar: str) -> None:
    """Checks that a tar file has the correct content structure and correct
    file name extension.

    :param tar: path to the tarfile
    :raises UserError:
    """
    if not tar.endswith(".tar"):
        raise UserError(
            f"Input file '{tar}' does not have .tar extension.\n"
            "Only .tar files can be used as input of the data "
            "decryption command."
        )
    check_tar_contents(os.path.realpath(tar))


def check_tar_contents(path: str) -> None:
    """Checks for exactly one .json file and exactly one .tar.gz.gpg

    :param path: path to the tarfile
    :raises UserError:
    """
    with tarfile.open(path, format=tarfile.PAX_FORMAT) as tar:
        content = set(tar.getnames())
        if content != PACKAGE_CONTENT:
            raise UserError(
                f"Input file '{path}' has non-compliant content:\n"
                + "\n".join(content)
                + "\n\n"
                "BioMedIT input tar files must contain exactly the following "
                "files:\n"
                f"\t- {METADATA_FILE} file with metadata\n"
                f"\t- {METADATA_FILE_SIG} file with a detached signature of the metadata\n"
                f"\t- {DATA_FILE_ENCRYPTED} file with the encrypted data"
            )


class ArchiveFileBase(ABC):
    def __init__(self, archive_path: str, content_container):
        self.archive_path = archive_path
        self.content_container = content_container

    @abstractmethod
    def add_to_tar(self, tar: tarfile.TarFile):
        pass


class ArchiveInMemoryFile(ArchiveFileBase):
    """An ArchiveFileBase whose content_container is a bytes instance."""

    def add_to_tar(self, tar: tarfile.TarFile):
        t = tarfile.TarInfo(self.archive_path)
        t.size = len(self.content_container)
        t.mtime = int(time.time())
        tar.addfile(t, io.BytesIO(self.content_container))


class ArchiveFile(ArchiveFileBase):
    """An ArchiveFileBase whose content_container is a file like object."""

    def add_to_tar(self, tar: tarfile.TarFile):
        t = tar.gettarinfo(
            name=self.content_container.name, arcname=self.archive_path, fileobj=None
        )
        with self.content_container as f_opened:
            tar.addfile(t, f_opened)


def write_tar(
    content: Sequence[ArchiveFileBase],
    output: Union[str, IO[bytes]],
    compress_level: int = 6,
    compress_algo: str = "gz",
) -> None:
    """Create a ".tar" or ".tar.gz" archive from files on disk or from a
    data stream, i.e. in-memory files.

    :param content: (in-memory) files to be added to the tar archive.
    :param output: path (or file object) or stream for the output tarball.
    :param compress_algo: compression algorithm to use when compressing
        the tar file (if compress_level > 0). Possible choices are
        'gz' (for gzip compression) and 'bz2' (for bzip2 compression).
    :param compress_level: gzip/bzip2 compression level in the range 0 (no
        compression) to 9 (highest compression).
    :raises UserError:
    """
    # Create a TypedDict class so that the mypy linter does not complain
    # about "Incompatible types in assignment".
    class TarfileArguments(TypedDict, total=False):
        name: str
        fileobj: IO[bytes]
        mode: str
        compresslevel: int

    archive_file_names = set(f.archive_path for f in content)
    assert_relative(archive_file_names)

    # Set correct mode depending on compression level and type of output,
    # i.e. file vs fileobject.
    # Note that currently the "w|gz" mode do not support the "compresslevel"
    # argument, so we use "w:gz" instead.
    tarfile_arguments: TarfileArguments = {
        "mode": f"w:{compress_algo}"
        if compress_level > 0
        else "w"
        if isinstance(output, str)
        else "w|"
    }
    if isinstance(output, str):
        tarfile_arguments["name"] = output
    else:
        tarfile_arguments["fileobj"] = output
    tarfile_arguments["mode"] = (
        f"w:{compress_algo}"
        if compress_level > 0
        else "w"
        if isinstance(output, str)
        else "w|"
    )
    if compress_level > 0:
        tarfile_arguments["compresslevel"] = compress_level

    # Compress input files into tarball.
    with tarfile.open(**tarfile_arguments) as tar:
        for archive_file in content:
            archive_file.add_to_tar(tar)

    if isinstance(output, str):
        # Verify the tarball file was created correctly. If not, delete it.
        if not Path(output).is_file():
            raise UserError(
                f"compression failed for [{output}]: " f"tarball not created."
            )
        with tarfile.open(output, "r") as tar:
            missing_files = {Path(p) for p in archive_file_names} - {
                Path(p) for p in tar.getnames()
            }
        if missing_files:
            delete_files(output)
            raise UserError(
                f"compression failed for [{output}]: "
                f"one or more files are missing in tarball."
            )


def is_relative(path: str):
    return not (os.path.isabs(path) or os.path.normpath(path).startswith(".."))


def assert_relative(files: Iterable[str]):
    """Asserts that each path in the list is relative"""
    non_relative = [f for f in files if not is_relative(f)]
    if non_relative:
        raise UserError(
            "The archive contains files with absolute path or paths ending "
            "in a parent directory:\n" + "\n".join(non_relative)
        )


def check_extracted(paths: Iterable[str]):
    """Tests if all paths in :paths: exist now on the file system"""
    failed = tuple(filterfalse(os.path.exists, paths))
    if failed:
        raise UserError("Failed to extract to following files: \n" + "\n".join(failed))


@contextmanager
def extract_with_progress(archive: IO[bytes], progress: ProgressInterface, *files):
    """Extracts given files from given archive.

    If only one file is specified, then progression scales with untaring of
    this specific file. Otherwise, each file takes an equal part (independent
    of its size) to progression.

    :param archive: the archive file.
    :param progress: the ProgressInterface. Must be specified.
    :param files: the list of tar members we want to untar.
    :raises UserError: if a 'KeyError' is thrown
    :returns: yields an io.BufferedReader or a casted 'FileObjectProgress'
    """
    with archive as a, tarfile.open(fileobj=a) as tar:
        try:
            if len(files) == 1:
                member = tar.extractfile(files[0])
                yield with_progress(member, progress) if member else None
            else:
                yield (
                    tar.extractfile(file_name)
                    for file_name in progress_iter(files, progress)
                )
        # 'tarfile.getmember' ('extractfile' invokes it): If 'file_name' can
        # not be found in the archive, 'KeyError' is raised.
        except KeyError as e:
            raise UserError(f"In archive {archive.name}: {e}") from e


@contextmanager
def extract(archive: IO[bytes], *files):
    with archive as a:
        with tarfile.open(fileobj=a) as tar:
            try:
                if len(files) == 1:
                    yield tar.extractfile(files[0])
                else:
                    yield (tar.extractfile(file_name) for file_name in files)

            except KeyError as e:
                raise UserError(f"In archive {archive.name}: {e.args[0]}") from e


def unpack_from_stream(
    archive: IO[bytes], dest: str, content: List[str], mode: str = "r|*"
):
    """Extract all files in the tarball from a stream.
    After unpacking the function checks that all files have been properly
    extracted by comparing the new files on disk to the list of files present
    in the tarball.

    :param archive: tarball stream
    :param dest: destination directory for the unpacked data
    :param content: a list container to store the file names compressed in
        the tarball
    :param mode: Mode to be passed to tarfile, i.e. `'r|*'`
    """
    if not os.path.isdir(dest):
        os.makedirs(dest)
    with tarfile.open(mode=mode, fileobj=archive) as tar:
        for f in tar:
            assert_relative((f.name,))
            content.append(f.name)
            tar.extract(f, dest)
    check_extracted((os.path.join(dest, f) for f in content))
